package ru.dbotthepony.mc.otm.matter;

import com.google.common.collect.ImmutableList;
import net.minecraft.core.BlockPos;
import net.minecraft.world.item.Item;
import net.minecraft.world.level.Level;
import net.minecraftforge.event.TickEvent;
import net.minecraftforge.eventbus.api.SubscribeEvent;
import net.minecraftforge.fmlserverevents.FMLServerStartedEvent;
import net.minecraftforge.fmlserverevents.FMLServerStoppingEvent;
import ru.dbotthepony.mc.otm.capability.*;

import javax.annotation.Nullable;
import java.math.BigDecimal;
import java.util.*;
import java.util.function.Consumer;
import java.util.function.Predicate;
import java.util.function.Supplier;

public class MatterGrid implements IMatterGridListener {
	public static final Set<MatterGrid> NETWORKS = new HashSet<>();

	private final Set<IMatterGridCell> cells = new HashSet<>();
	private final Set<IMatterGridListener> listeners = new HashSet<>();

	private static final Set<Supplier<Boolean>> discovering_neighbours = new HashSet<>();

	public static void scheduleDiscoverNeighbours(IMatterGridCell cell, BlockPos pos, Level level) {
		discovering_neighbours.add(() -> !cell.isValidMatterCell() || cell.connectOrCreateMatterGrid(pos, level, true));
	}

	@SubscribeEvent
	public void onServerStarting(FMLServerStartedEvent event) {
		discovering_neighbours.clear();
	}

	@SubscribeEvent
	public void onServerStopping(FMLServerStoppingEvent event) {
		discovering_neighbours.clear();
	}

	@SubscribeEvent
	public static void discoverNeighbours(TickEvent.ServerTickEvent event) {
		if (event.phase == TickEvent.Phase.END && discovering_neighbours.size() != 0) {
			ArrayList<Supplier<Boolean>> invalid = new ArrayList<>();

			for (Supplier<Boolean> cell : discovering_neighbours) {
				if (cell.get()) {
					invalid.add(cell);
				}
			}

			if (invalid.size() != 0) {
				for (Supplier<Boolean> cell : invalid) {
					discovering_neighbours.remove(cell);
				}
			}
		}
	}

	public MatterGrid() {
		NETWORKS.add(this);
	}

	public BigDecimal getCapacity() {
		BigDecimal summ = BigDecimal.ZERO;

		for (IMatterGridCell cell : cells) {
			IMatterHandler handler = cell.getMatterHandler();

			if (handler != null && handler.getDirection() == IMatterHandler.MatterDirection.BIDIRECTIONAL) {
				summ = summ.add(handler.getMaxStoredMatter());
			}
		}

		return summ;
	}

	public BigDecimal getPotentialCapacity() {
		BigDecimal summ = BigDecimal.ZERO;

		for (IMatterGridCell cell : cells) {
			IMatterHandler handler = cell.getMatterHandler();

			if (handler != null && handler.getDirection() != IMatterHandler.MatterDirection.RECEIVE) {
				summ = summ.add(handler.getMaxStoredMatter());
			}
		}

		return summ;
	}

	public BigDecimal getStored() {
		BigDecimal summ = BigDecimal.ZERO;

		for (IMatterGridCell cell : cells) {
			IMatterHandler handler = cell.getMatterHandler();

			if (handler != null && handler.getDirection() == IMatterHandler.MatterDirection.BIDIRECTIONAL) {
				summ = summ.add(handler.getStoredMatter());
			}
		}

		return summ;
	}

	public BigDecimal getPotentialStored() {
		BigDecimal summ = BigDecimal.ZERO;

		for (IMatterGridCell cell : cells) {
			IMatterHandler handler = cell.getMatterHandler();

			if (handler != null && handler.getDirection() != IMatterHandler.MatterDirection.RECEIVE) {
				summ = summ.add(handler.getStoredMatter());
			}
		}

		return summ;
	}

	public BigDecimal extractMatter(BigDecimal howMuch, boolean simulate) {
		if (howMuch.compareTo(BigDecimal.ZERO) <= 0)
			return BigDecimal.ZERO;

		validate();

		BigDecimal extracted = BigDecimal.ZERO;

		for (IMatterGridCell cell : cells) {
			IMatterHandler handler = cell.getMatterHandler();

			if (handler != null && handler.getDirection() != IMatterHandler.MatterDirection.RECEIVE) {
				BigDecimal drain = handler.extractMatterOuter(howMuch, simulate);

				if (!drain.equals(BigDecimal.ZERO)) {
					extracted = extracted.add(drain);
					howMuch = howMuch.subtract(drain);

					if (howMuch.compareTo(BigDecimal.ZERO) <= 0) {
						break;
					}
				}
			}
		}

		return extracted;
	}

	public BigDecimal softPushMatter(BigDecimal howMuch, boolean simulate) {
		if (howMuch.compareTo(BigDecimal.ZERO) <= 0)
			return BigDecimal.ZERO;

		validate();

		BigDecimal received = BigDecimal.ZERO;

		for (IMatterGridCell cell : cells) {
			IMatterHandler handler = cell.getMatterHandler();

			if (handler != null && handler.getDirection() == IMatterHandler.MatterDirection.BIDIRECTIONAL) {
				BigDecimal receive = handler.receiveMatterOuter(howMuch, simulate);

				if (!receive.equals(BigDecimal.ZERO)) {
					received = received.add(receive);
					howMuch = howMuch.subtract(receive);

					if (howMuch.compareTo(BigDecimal.ZERO) <= 0) {
						break;
					}
				}
			}
		}

		return received;
	}

	public BigDecimal pushMatter(BigDecimal howMuch, boolean simulate) {
		if (howMuch.compareTo(BigDecimal.ZERO) <= 0)
			return BigDecimal.ZERO;

		validate();

		BigDecimal received = BigDecimal.ZERO;

		for (IMatterGridCell cell : cells) {
			IMatterHandler handler = cell.getMatterHandler();

			if (handler != null && handler.getDirection() != IMatterHandler.MatterDirection.EXTRACT) {
				BigDecimal receive = handler.receiveMatterOuter(howMuch, simulate);

				if (!receive.equals(BigDecimal.ZERO)) {
					received = received.add(receive);
					howMuch = howMuch.subtract(receive);

					if (howMuch.compareTo(BigDecimal.ZERO) <= 0) {
						break;
					}
				}
			}
		}

		return received;
	}

	public PatternInsertStatus insertPattern(PatternState state, boolean only_update, boolean simulate) {
		validate();

		for (IMatterGridCell cell : cells) {
			IPatternStorage storage = cell.getPatternStorage();

			if (storage != null) {
				var status = storage.insertPattern(state, true, simulate);

				if (status.status() != PatternInsertStatus.Status.FAIL) {
					return status;
				}
			}
		}

		if (only_update)
			return new PatternInsertStatus();

		for (IMatterGridCell cell : cells) {
			IPatternStorage storage = cell.getPatternStorage();

			if (storage != null) {
				var status = storage.insertPattern(state, false, simulate);

				if (status.status() != PatternInsertStatus.Status.FAIL) {
					return status;
				}
			}
		}

		return new PatternInsertStatus();
	}

	public Collection<PatternState> getStoredPatterns() {
		validate();

		ArrayList<PatternState> list = new ArrayList<>();

		for (IMatterGridCell cell : cells) {
			IPatternStorage storage = cell.getPatternStorage();

			if (storage != null) {
				list.addAll(storage.getStoredPatterns());
			}
		}

		return ImmutableList.copyOf(list);
	}

	public Collection<PatternState> findPatterns(Item item) {
		return findPatterns((patternState -> item == patternState.item()));
	}

	public Collection<PatternState> findPatterns(Predicate<PatternState> item) {
		validate();

		ArrayList<PatternState> list = new ArrayList<>();

		for (IMatterGridCell cell : cells) {
			IPatternStorage storage = cell.getPatternStorage();

			if (storage != null) {
				list.addAll(storage.findPatterns(item));
			}
		}

		return ImmutableList.copyOf(list);
	}

	@Nullable
	public PatternState getPattern(PatternState state) {
		return getPattern(state.id());
	}

	@Nullable
	public PatternState getPattern(UUID id) {
		validate();

		for (IMatterGridCell cell : cells) {
			IPatternStorage storage = cell.getPatternStorage();

			if (storage != null) {
				var get = storage.getPattern(id);

				if (get != null) {
					return get;
				}
			}
		}

		return null;
	}

	public boolean hasPattern(UUID id) {
		return getPattern(id) != null;
	}

	@Nullable
	public MatterTaskAllocation allocateTask(boolean simulate) {
		validate();

		for (IMatterGridCell cell : cells) {
			IMatterTaskProvider tasks = cell.getTaskProvider();

			if (tasks != null) {
				MatterTaskAllocation allocated = tasks.allocateTask(simulate);

				if (allocated != null) {
					return allocated;
				}
			}
		}

		return null;
	}

	@Nullable
	public boolean notifyTaskCompletion(MatterTask task) {
		validate();

		for (IMatterGridCell cell : cells) {
			IMatterTaskProvider tasks = cell.getTaskProvider();

			if (tasks != null) {
				if (tasks.notifyTaskCompletion(task)) {
					return true;
				}
			}
		}

		return false;
	}

	public void track(IMatterGridCell entity) {
		if (cells.contains(entity))
			return;

		cells.add(entity);
		entity.setMatterGrid(this);

		var tasks = entity.getTaskProvider();

		if (tasks != null) {
			for (var task : tasks.getAllTasks()) {
				onMatterTaskCreated(task);
			}
		}

		var patterns = entity.getPatternStorage();

		if (patterns != null) {
			for (var pattern : patterns.getStoredPatterns()) {
				onPatternAdded(pattern);
			}
		}
	}

	public void untrack(IMatterGridCell entity) {
		if (!cells.contains(entity))
			return;

		cells.remove(entity);
		entity.setMatterGrid(null);

		var tasks = entity.getTaskProvider();

		if (tasks != null) {
			for (var task : tasks.getAllTasks()) {
				onMatterTaskRemoved(task);
			}
		}

		var patterns = entity.getPatternStorage();

		if (patterns != null) {
			for (var pattern : patterns.getStoredPatterns()) {
				onPatternRemoved(pattern);
			}
		}

		if (cells.size() == 0)
			NETWORKS.remove(this);
	}

	public void attach(IMatterGridListener listener) {
		listeners.add(listener);
	}

	public void detach(IMatterGridListener listener) {
		listeners.remove(listener);
	}

	public int networkSize() {
		return cells.size();
	}

	public void mergeWith(MatterGrid other) {
		if (other == this)
			return;

		MatterGrid a;
		MatterGrid b;

		if (networkSize() > other.networkSize()) {
			a = this;
			b = other;
		} else {
			a = other;
			b = this;
		}

		b.validate();

		for (IMatterGridCell entity : b.cells) {
			a.track(entity);
		}

		NETWORKS.remove(b);
	}

	public void validate() {
		ArrayList<IMatterGridCell> invalid = new ArrayList<>();

		for (IMatterGridCell entity : cells) {
			if (!entity.isValidMatterCell()) {
				invalid.add(entity);
			}
		}

		if (invalid.size() != 0) {
			for (IMatterGridCell entity : invalid) {
				untrack(entity);
			}
		}

		if (cells.size() == 0) {
			NETWORKS.remove(this);
		}
	}

	@Override
	public void onPatternAdded(PatternState state) {
		// validate();
		cells.forEach(cell -> cell.onPatternAdded(state));
		listeners.forEach(cell -> cell.onPatternAdded(state));
	}

	@Override
	public void onPatternRemoved(PatternState state) {
		// validate();
		cells.forEach(cell -> cell.onPatternRemoved(state));
		listeners.forEach(cell -> cell.onPatternRemoved(state));
	}

	@Override
	public void onPatternUpdated(PatternState new_state, PatternState old_state) {
		// validate();
		cells.forEach(cell -> cell.onPatternUpdated(new_state, old_state));
		listeners.forEach(cell -> cell.onPatternUpdated(new_state, old_state));
	}

	@Override
	public void onMatterTaskCreated(MatterTask task) {
		// validate();
		cells.forEach(cell -> cell.onMatterTaskCreated(task));
		listeners.forEach(cell -> cell.onMatterTaskCreated(task));
	}

	@Override
	public void onMatterTaskUpdated(MatterTask new_state, MatterTask old_state) {
		// validate();
		cells.forEach(cell -> cell.onMatterTaskUpdated(new_state, old_state));
		listeners.forEach(cell -> cell.onMatterTaskUpdated(new_state, old_state));
	}

	@Override
	public void onMatterTaskFinished(MatterTask state) {
		// validate();
		cells.forEach(cell -> cell.onMatterTaskFinished(state));
		listeners.forEach(cell -> cell.onMatterTaskFinished(state));
	}

	@Override
	public void onMatterTaskRemoved(MatterTask state) {
		// validate();
		cells.forEach(cell -> cell.onMatterTaskRemoved(state));
		listeners.forEach(cell -> cell.onMatterTaskRemoved(state));
	}
}
